"""
The module that provides generating methods for all parts
of the pdelab driver.

Currently, these are hardcoded as strings. It would be possible
to switch these to cgen expression. OTOH, there is not much to be
gained there.
"""
from dune.perftool.generation import (generator_factory,
                                      include_file,
                                      symbol,
                                      preamble,
                                      )
from dune.perftool.options import get_option

# Have a global variable with the entire form data. This allows functions that depend
# deterministically on the entire data set to directly access it instead of passing it
# through the entire generator chain.
_driver_data = {}


# Have a function access this global data structure
def set_driver_data(formdatas, data):
    assert (len(formdatas) <= 2)
    if len(formdatas) == 1:
        _driver_data['form'] = formdatas[0].preprocessed_form
        _driver_data['formdata'] = formdatas[0]
    else:
        mass_index = mass_form_index(formdatas, data)
        if mass_index is None:
            raise NotImplementedError("Form for mass matrix needs to have name 'mass' in ufl file.")
        _driver_data['mass_form'] = formdatas[mass_index].preprocessed_form
        _driver_data['mass_formdata'] = formdatas[mass_index]
        _driver_data['form'] = formdatas[1 - mass_index].preprocessed_form
        _driver_data['formdata'] = formdatas[1 - mass_index]

    _driver_data['data'] = data


def is_stationary():
    return 'mass_form' not in _driver_data


def form_name_suffix(name, formdata):
    from dune.perftool.pdelab.localoperator import name_form
    data = _driver_data['data']
    form_name = name_form(formdata, data)
    return name + '_' + form_name


def has_constraints(element):
    from ufl import MixedElement, VectorElement
    if isinstance(element, MixedElement) and not isinstance(element, VectorElement):
        return any(has_constraints(el) for el in element.sub_elements())
    return hasattr(element, "dirichlet_constraints")


# Those symbol generators that depend on the meta data we hacked into
# the ufl.FiniteElement should use fem_metadata_dependent_symbol instead of symbol.
def get_constraints_metadata(element):
    """
    Normally, we suppress the constraints information on a finite element.
    However, we might want to use it explicitly in those case where it matters
    This function does the job!
    """
    dc = getattr(element, 'dirichlet_constraints', None)
    de = getattr(element, 'dirichlet_expression', None)
    return (element, (dc, de))


fem_metadata_dependent_symbol = generator_factory(item_tags=("symbol",),
                                                  cache_key_generator=get_constraints_metadata,
                                                  )


def mass_form_index(formdatas, data):
    for index, formdata in enumerate(formdatas):
        try:
            if data.object_names[id(formdata.original_form)] == 'mass':
                return index
        except KeyError:
            continue


def is_linear(form):
    '''Test if form is linear in test function'''
    from ufl import derivative
    from ufl.algorithms import expand_derivatives
    jacform = expand_derivatives(derivative(form, form.coefficients()[0]))
    for coeff in jacform.coefficients():
        if 0 == coeff.count():
            return False

    return True


def isLagrange(fem):
    return fem._short_name is 'CG'


def isSimplical(fem):
    return any(fem._cell._cellname in x for x in ["triangle", "tetrahedron"])


def isQuadrilateral(fem):
    return any(fem._cell._cellname in x for x in ["vertex", "interval", "quadrilateral", "hexahedron"])


def isPk(fem):
    return isLagrange(fem) and isSimplical(fem)


def isQk(fem):
    return isLagrange(fem) and isQuadrilateral(fem)


def isDG(fem):
    return fem._short_name is 'DG'


def FEM_name_mangling(fem):
    from ufl import MixedElement, VectorElement, FiniteElement, TensorElement
    if isinstance(fem, VectorElement):
        return FEM_name_mangling(fem.sub_elements()[0]) + "_" + str(fem.num_sub_elements())
    if isinstance(fem, TensorElement):
        return FEM_name_mangling(fem.sub_elements()[0]) + "_" + "_".join(str(i) for i in fem.value_shape())
    if isinstance(fem, MixedElement):
        name = ""
        for elem in fem.sub_elements():
            if name is not "":
                name = name + "_"
            name = name + FEM_name_mangling(elem)
        return name
    if isinstance(fem, FiniteElement):
        if isPk(fem):
            return "P" + str(fem._degree)
        if isQk(fem):
            return "Q" + str(fem._degree)
        if isDG(fem):
            return "DG" + str(fem._degree)

    raise NotImplementedError("FEM NAME MANGLING")


@symbol
def name_inifile():
    # TODO pass some other option here.
    return "argv[1]"


@preamble
def parse_initree(varname):
    include_file("dune/common/parametertree.hh", filetag="driver")
    include_file("dune/common/parametertreeparser.hh", filetag="driver")
    filename = name_inifile()
    return ["Dune::ParameterTree initree;", "Dune::ParameterTreeParser::readINITree({}, {});".format(filename, varname)]


@symbol
def name_initree():
    parse_initree("initree")
    # TODO we can get some other ini file here.
    return "initree"


@preamble
def define_dimension(name):
    return "static const int {} = {};".format(name, _driver_data['form'].ufl_cell().geometric_dimension())


@symbol
def name_dimension():
    define_dimension("dim")
    return "dim"


@preamble
def typedef_grid(name):
    dim = name_dimension()
    if any(_driver_data['form'].ufl_cell().cellname() in x for x in ["vertex", "interval", "quadrilateral", "hexahedron"]):
        # For Yasp Grids the jacobi of the transformation is diagonal
        from dune.perftool.options import set_option
        set_option('diagonal_transformation_matrix', True)

        gridt = "Dune::YaspGrid<{}>".format(dim)
        include_file("dune/grid/yaspgrid.hh", filetag="driver")
    else:
        if any(_driver_data['form'].ufl_cell().cellname() in x for x in ["triangle", "tetrahedron"]):
            # gridt = "Dune::UGGrid<{}>".format(dim)
            # include_file("dune/grid/uggrid.hh", filetag="driver")
            gridt = "Dune::ALUGrid<{}, {}, Dune::simplex, Dune::conforming>".format(dim, dim)
            include_file("dune/alugrid/grid.hh", filetag="driver")
        else:
            raise ValueError("Cant match your geometry with a DUNE grid. Please report bug.")
    return "typedef {} {};".format(gridt, name)


@symbol
def type_grid():
    typedef_grid("Grid")
    return "Grid"


@preamble
def define_grid(name):
    include_file("dune/testtools/gridconstruction.hh", filetag="driver")
    ini = name_initree()
    _type = type_grid()
    return ["IniGridFactory<{}> factory({});".format(_type, ini),
            "std::shared_ptr<{}> grid = factory.getGrid();".format(_type)]


@symbol
def name_grid():
    define_grid("grid")
    return "grid"


@preamble
def typedef_leafview(name):
    grid = type_grid()
    return "typedef {}::LeafGridView {};".format(grid, name)


@symbol
def type_leafview():
    typedef_leafview("GV")
    return "GV"


@preamble
def define_leafview(name):
    _type = type_leafview()
    grid = name_grid()
    return "{} {} = {}->leafGridView();".format(_type, name, grid)


@symbol
def name_leafview():
    define_leafview("gv")
    return "gv"


@preamble
def typedef_vtkwriter(name):
    include_file("dune/grid/io/file/vtk/subsamplingvtkwriter.hh", filetag="driver")
    gv = type_leafview()
    return "typedef Dune::SubsamplingVTKWriter<{}> {};".format(gv, name)


@symbol
def type_vtkwriter():
    typedef_vtkwriter("VTKWriter")
    return "VTKWriter"


@preamble
def define_subsamplinglevel(name):
    ini = name_initree()
    return "int {} = {}.get<int>(\"vtk.subsamplinglevel\", 0);".format(name, ini)


@symbol
def name_subsamplinglevel():
    define_subsamplinglevel("sublevel")
    return "sublevel"


@preamble
def define_vtkwriter(name):
    _type = type_vtkwriter()
    gv = name_leafview()
    subsamp = name_subsamplinglevel()
    return "{} {}({}, {});".format(_type, name, gv, subsamp)


@symbol
def name_vtkwriter():
    define_vtkwriter("vtkwriter")
    return "vtkwriter"


@preamble
def typedef_domainfield(name):
    gridt = type_grid()
    return "typedef {}::ctype {};".format(gridt, name)


@symbol
def type_domainfield():
    typedef_domainfield("DF")
    return "DF"


@preamble
def typedef_range(name):
    return "typedef double {};".format(name)


@symbol
def type_range():
    typedef_range("R")
    return "R"


@preamble
def typedef_fem(expr, name):
    gv = type_leafview()
    df = type_domainfield()
    r = type_range()
    dim = name_dimension()
    if isPk(expr):
        include_file("dune/pdelab/finiteelementmap/pkfem.hh", filetag="driver")
        return "typedef Dune::PDELab::PkLocalFiniteElementMap<{}, {}, {}, {}> {};".format(gv, df, r, expr._degree, name)
    if isQk(expr):
        include_file("dune/pdelab/finiteelementmap/qkfem.hh", filetag="driver")
        return "typedef Dune::PDELab::QkLocalFiniteElementMap<{}, {}, {}, {}> {};".format(gv, df, r, expr._degree, name)
    if isDG(expr):
        if isQuadrilateral(expr):
            include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="driver")
            # TODO allow switching the basis here!
            return "typedef Dune::PDELab::QkDGLocalFiniteElementMap<{}, {}, {}, {}> {};".format(df, r, expr._degree, dim, name)
        if isSimplical(expr):
            include_file("dune/pdelab/finiteelementmap/opbfem.hh", filetag="driver")
            return "typedef Dune::PDELab::OPBLocalFiniteElementMap<{}, {}, {}, {}, Dune::GeometryType::simplex> {};".format(df, r, expr._degree, dim, name)
        raise NotImplementedError("Geometry type not known in code generation")

    raise NotImplementedError("FEM not implemented in dune-perftool")


@symbol
def type_fem(expr):
    name = "{}_FEM".format(FEM_name_mangling(expr).upper())
    typedef_fem(expr, name)
    return name


@preamble
def define_fem(expr, name):
    femtype = type_fem(expr)
    if isDG(expr):
        return "{} {};".format(femtype, name)
    else:
        gv = name_leafview()
        return "{} {}({});".format(femtype, name, gv)


@symbol
def name_fem(expr):
    name = "{}_fem".format(FEM_name_mangling(expr).lower())
    define_fem(expr, name)
    return name


@preamble
def typedef_vectorbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::istl::VectorBackend<Dune::PDELab::istl::Blocking::none, 1> {};".format(name)


@symbol
def type_vectorbackend():
    typedef_vectorbackend("VectorBackend")
    return "VectorBackend"


@symbol
def type_orderingtag():
    return "Dune::PDELab::LexicographicOrderingTag"


@preamble
def typedef_dirichlet_constraintsassembler(name):
    include_file("dune/pdelab/constraints/conforming.hh", filetag="driver")
    return "typedef Dune::PDELab::ConformingDirichletConstraints {};".format(name)


@preamble
def typedef_no_constraintsassembler(name):
    return "typedef Dune::PDELab::NoConstraints {};".format(name)


@symbol
def type_constraintsassembler(dirichlet):
    if dirichlet[1]:
        typedef_dirichlet_constraintsassembler("DirichletConstraintsAssember")
        return "DirichletConstraintsAssember"
    else:
        typedef_no_constraintsassembler("NoConstraintsAssembler")
        return "NoConstraintsAssembler"


@preamble
def typedef_constraintscontainer(element, name):
    gfs = type_gfs(element)
    r = type_range()
    return "typedef {}::ConstraintsContainer<{}>::Type {};".format(gfs, r, name)


@symbol
def type_constraintscontainer(element):
    name = "{}_cc".format(FEM_name_mangling(element)).upper()
    typedef_constraintscontainer(element, name)
    return name


@preamble
def define_constraintscontainer(expr, name):
    cctype = type_constraintscontainer(expr)
    return ["{} {};".format(cctype, name), "{}.clear();".format(name)]


# TODO this does not yet fully support having same elements with different constraints!
@fem_metadata_dependent_symbol
def name_constraintscontainer(expr):
    name = "{}{}_cc".format(FEM_name_mangling(expr), "_dirichlet" if has_constraints(expr) else "").lower()
    define_constraintscontainer(expr, name)
    return name


@preamble
def define_intersection_lambda(expression, name):
    if expression is None:
        return "auto {} = [&](const auto& x){{ return 0; }};".format(name)
    if expression.is_global:
        return "auto {} = [&](const auto& x){{ {} }};".format(name, expression.c_expr)
    else:
        return "auto {} = [&](const auto& is, const auto& x){{ {} }};".format(name, expression.c_expr)


@symbol
def name_bctype_lambda(name, dirichlet):
    name = name + "_lambda"
    define_intersection_lambda(dirichlet, name)
    return name


@preamble
def define_bctype_function(dirichlet, name):
    gv = name_leafview()
    bctype_lambda = name_bctype_lambda(name, dirichlet)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    return "auto {} = Dune::PDELab::makeBoundaryConditionFromCallable({}, {});".format(name,
                                                                                       gv,
                                                                                       bctype_lambda,
                                                                                       )


@preamble
def define_powergfs_constraints(name, child, dim):
    include_file('dune/pdelab/constraints/common/constraintsparameters.hh', filetag='driver')
    return "Dune::PDELab::PowerConstraintsParameters<decltype({}), {}> {}({});".format(child, dim, name, child)


@preamble
def define_compositegfs_constraints(name, *children):
    include_file('dune/pdelab/constraints/common/constraintsparameters.hh', filetag='driver')
    return "Dune::PDELab::CompositeConstraintsParameters<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in children),
                                                                             name,
                                                                             ', '.join(c for c in children)
                                                                             )


@symbol
def name_bctype_function(expr):
    from ufl import MixedElement, VectorElement, TensorElement
    if isinstance(expr, (VectorElement, TensorElement)):
        element, (dirichlet, _) = get_constraints_metadata(expr)
        # get the correct name from the corresponding UFL file
        from dune.perftool.generation import get_global_context_value
        name = get_global_context_value("data").object_names.get(id(dirichlet), "everywhere")
        define_bctype_function(dirichlet, name)
        pgfs_name = '{}_{}'.format(name, expr.num_sub_elements())
        define_powergfs_constraints(pgfs_name, name, expr.num_sub_elements())
        return pgfs_name
    if isinstance(expr, MixedElement):
        children = tuple(name_bctype_function(el) for el in expr.sub_elements())
        name = '_'.join(children)
        define_compositegfs_constraints(name, *children)
        return name

    # So, this seems to be a leaf!
    element, (dirichlet, _) = get_constraints_metadata(expr)

    # get the correct name from the corresponding UFL file
    from dune.perftool.generation import get_global_context_value
    name = get_global_context_value("data").object_names.get(id(dirichlet), "everywhere")

    define_bctype_function(dirichlet, name)
    return name


@preamble
def assemble_constraints(name, expr):
    bctype_function = name_bctype_function(expr)
    gfs = name_gfs(expr)
    return "Dune::PDELab::constraints({}, {}, {});".format(bctype_function,
                                                           gfs,
                                                           name,
                                                           )


@symbol
def name_assembled_constraints(expr):
    name = name_constraintscontainer(expr)
    if has_constraints(expr):
        assemble_constraints(name, expr)
    return name


@preamble
def typedef_gfs(element, dirichlet, name):
    vb = type_vectorbackend()
    from ufl import FiniteElement, MixedElement, VectorElement, EnrichedElement, RestrictedElement, TensorElement
    if isinstance(element, FiniteElement):
        gv = type_leafview()
        fem = type_fem(element)
        cass = type_constraintsassembler(dirichlet)
        return "typedef Dune::PDELab::GridFunctionSpace<{}, {}, {}, {}> {};".format(gv, fem, cass, vb, name)
    if isinstance(element, (VectorElement, TensorElement)):
        include_file("dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh", filetag="driver")
        gv = type_leafview()
        fem = type_fem(element.sub_elements()[0])
        # TODO implement symmetry here
        dim = element.num_sub_elements()
        cass = type_constraintsassembler(dirichlet)
        return "typedef Dune::PDELab::VectorGridFunctionSpace<{}, {}, {}, {}, {}, {}> {};".format(gv, fem, dim, vb, vb, cass, name)
    if isinstance(element, MixedElement):
        ot = type_orderingtag()
        args = ", ".join(type_gfs(e) for e in element.sub_elements())
        return "typedef Dune::PDELab::CompositeGridFunctionSpace<{}, {}, {}> {};".format(vb, ot, args, name)
    if isinstance(element, EnrichedElement):
        raise NotImplementedError("Dune does not support enriched elements!")
    if isinstance(element, RestrictedElement):
        raise NotImplementedError("Dune does not support restricted elements!")


@fem_metadata_dependent_symbol
def type_gfs(expr):
    element, dirichlet = get_constraints_metadata(expr)
    name = "{}{}_gfs".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").upper()
    typedef_gfs(element, dirichlet, name)
    return name


@preamble
def define_gfs(element, dirichlet, name):
    gfstype = type_gfs(element)
    from ufl import FiniteElement, MixedElement, VectorElement, EnrichedElement, RestrictedElement, TensorElement
    if isinstance(element, FiniteElement):
        gv = name_leafview()
        fem = name_fem(element)
        return "{} {}({}, {});".format(gfstype, name, gv, fem)
    if isinstance(element, (VectorElement, TensorElement)):
        gv = name_leafview()
        fem = name_fem(element.sub_elements()[0])
        return "{} {}({}, {});".format(gfstype, name, gv, fem)
    if isinstance(element, MixedElement):
        args = ", ".join(name_gfs(childgfs) for childgfs in element.sub_elements())
        return "{} {}({});".format(gfstype, name, args)
    raise NotImplementedError("Dune does not support this type of finite element!")


@fem_metadata_dependent_symbol
def name_gfs(expr):
    element, dirichlet = get_constraints_metadata(expr)
    name = "{}{}_gfs".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").lower()
    define_gfs(element, dirichlet, name)
    return name


@preamble
def define_dofestimate(name):
    # Provide a worstcase estimate for the number of entries per row based on the given gridfunction space and cell geometry
    if isQuadrilateral(_driver_data['form'].coefficients()[0].ufl_element()):
        geo_factor = "4"
    else:
        geo_factor = "6"
    gfs = name_gfs(_driver_data['form'].coefficients()[0].ufl_element())
    ini = name_initree()
    return ["int generic_dof_estimate =  {} * {}.maxLocalSize();".format(geo_factor, gfs),
            "int {} = {}.get<int>(\"istl.number_of_nnz\", generic_dof_estimate);".format(name, ini)]


@symbol
def name_dofestimate():
    define_dofestimate("dofestimate")
    return "dofestimate"


@preamble
def typedef_matrixbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::istl::BCRSMatrixBackend<> {};".format(name)


@symbol
def type_matrixbackend():
    typedef_matrixbackend("MatrixBackend")
    return "MatrixBackend"


@preamble
def define_matrixbackend(name):
    mbtype = type_matrixbackend()
    dof = name_dofestimate()
    return "{} {}({});".format(mbtype, name, dof)


@symbol
def name_matrixbackend():
    define_matrixbackend("mb")
    return "mb"


@symbol
def type_parameters(formdata):
    from dune.perftool.generation import get_global_context_value
    data = get_global_context_value("data")
    from dune.perftool.pdelab.parameter import parameterclass_basename
    name = parameterclass_basename(formdata, data)
    return name


@preamble
def define_parameters(name, formdata):
    partype = type_parameters(formdata)
    return "{} {};".format(partype, name)


@symbol
def name_parameters(formdata):
    name = form_name_suffix("params", formdata)
    define_parameters(name, formdata)
    return name


@preamble
def typedef_localoperator(name, formdata):
    # No Parameter class here, yet
    # params = type_parameters()
    # return "typedef LocalOperator<{}> {};".format(params, name)
    ugfs = type_gfs(_driver_data['form'].coefficients()[0].ufl_element())
    vgfs = type_gfs(_driver_data['form'].arguments()[0].ufl_element())
    from dune.perftool.generation import get_global_context_value
    data = get_global_context_value("data")
    filename = get_option("operator_file")
    include_file(filename, filetag="driver")
    from dune.perftool.pdelab.localoperator import localoperator_basename
    lopname = localoperator_basename(formdata, data)
    return "using {} = {}<{}, {}>;".format(name, lopname, ugfs, vgfs)


@symbol
def type_localoperator(formdata):
    name = form_name_suffix("LOP", formdata).upper()
    typedef_localoperator(name, formdata)
    return name


@preamble
def define_localoperator(name, formdata):
    loptype = type_localoperator(formdata)
    ini = name_initree()
    params = name_parameters(formdata)
    return "{} {}({}, {});".format(loptype, name, ini, params)


@symbol
def name_localoperator(formdata):
    name = form_name_suffix("lop", formdata)
    define_localoperator(name, formdata)
    return name


@preamble
def typedef_gridoperator(name, formdata):
    ugfs = type_gfs(_driver_data['form'].coefficients()[0].ufl_element())
    vgfs = type_gfs(_driver_data['form'].arguments()[0].ufl_element())
    lop = type_localoperator(formdata)
    ucc = type_constraintscontainer(_driver_data['form'].coefficients()[0].ufl_element())
    vcc = type_constraintscontainer(_driver_data['form'].arguments()[0].ufl_element())
    mb = type_matrixbackend()
    df = type_domainfield()
    r = type_range()
    include_file("dune/pdelab/gridoperator/gridoperator.hh", filetag="driver")
    return "typedef Dune::PDELab::GridOperator<{}, {}, {}, {}, {}, {}, {}, {}, {}> {};".format(ugfs, vgfs, lop, mb, df, r, r, ucc, vcc, name)


@symbol
def type_gridoperator(formdata):
    name = form_name_suffix("GO", formdata).upper()
    typedef_gridoperator(name, formdata)
    return name


@preamble
def define_gridoperator(name, formdata):
    gotype = type_gridoperator(formdata)
    # TODO use formdata insteat of _driver_data object
    ugfs = name_gfs(_driver_data['form'].coefficients()[0].ufl_element())
    ucc = name_assembled_constraints(_driver_data['form'].coefficients()[0].ufl_element())
    vgfs = name_gfs(_driver_data['form'].arguments()[0].ufl_element())
    vcc = name_assembled_constraints(_driver_data['form'].arguments()[0].ufl_element())
    lop = name_localoperator(formdata)
    mb = name_matrixbackend()
    return ["{} {}({}, {}, {}, {}, {}, {});".format(gotype, name, ugfs, ucc, vgfs, vcc, lop, mb),
            "std::cout << \"gfs with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ugfs),
            "std::cout << \"cc with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ucc)]


@symbol
def name_gridoperator(formdata):
    name = form_name_suffix("go", formdata)
    define_gridoperator(name, formdata)
    return name


@preamble
def typedef_vector(name, formdata):
    go_type = type_gridoperator(formdata)
    return "typedef {}::Traits::Domain {};".format(go_type, name)


@symbol
def type_vector(formdata):
    name = form_name_suffix("V", formdata).upper()
    typedef_vector(name, formdata)
    return name


@preamble
def define_vector(name, formdata):
    vtype = type_vector(formdata)
    gfs = name_gfs(_driver_data['form'].coefficients()[0].ufl_element())
    return ["{} {}({});".format(vtype, name, gfs), "{} = 0.0;".format(name)]


@preamble
def define_boundary_lambda(boundary, name):
    if boundary is None:
        return "auto {} = [&](const auto& x){{ return 0.0; }};".format(name)
    if boundary.is_global:
        return "auto {} = [&](const auto& x){{ {} }};".format(name, boundary.c_expr)
    else:
        return "auto {} = [&](const auto& e, const auto& x){{ {} }};".format(name, boundary.c_expr)


@symbol
def name_boundary_lambda(boundary, name):
    define_boundary_lambda(boundary, name + "lambda")
    return name + "lambda"


@preamble
def define_boundary_function(boundary, name):
    gv = name_leafview()
    lambdaname = name_boundary_lambda(boundary, name)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    if is_stationary():
        return "auto {} = Dune::PDELab::makeGridFunctionFromCallable({}, {});".format(name,
                                                                                      gv,
                                                                                      lambdaname,
                                                                                      )
    else:
        params = name_parameters(_driver_data['formdata'])
        return "auto {} = Dune::PDELab::makeInstationaryGridFunctionFromCallable({}, {}, {});".format(name,
                                                                                                      gv,
                                                                                                      lambdaname,
                                                                                                      params)


@preamble
def define_compositegfs_parameterfunction(name, *children):
    return "Dune::PDELab::CompositeGridFunction<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in children),
                                                                    name,
                                                                    ', '.join(children))


@symbol
def name_boundary_function(expr):
    from ufl import MixedElement, VectorElement, TensorElement
    if isinstance(expr, (VectorElement, TensorElement)):
        element, (_, boundary) = get_constraints_metadata(expr)
        from dune.perftool.generation import get_global_context_value
        name = get_global_context_value("data").object_names.get(id(boundary), "zero")
        define_boundary_function(boundary, name)
        return name
    if isinstance(expr, MixedElement):
        children = tuple(name_boundary_function(el) for el in expr.sub_elements())
        name = '_'.join(children)
        define_compositegfs_parameterfunction(name, *children)
        return name

    # This is a scalar leaf of the ansatz function tree
    element, (_, boundary) = get_constraints_metadata(expr)

    from dune.perftool.generation import get_global_context_value
    name = get_global_context_value("data").object_names.get(id(boundary), "zero")

    define_boundary_function(boundary, name)
    return name


@symbol
def name_solution_function(expr):
    # Get expression representing solution
    from dune.perftool.generation import get_global_context_value
    solution_expression_name = get_option("exact_solution_expression")
    solution_expression = get_global_context_value("data").object_by_name[solution_expression_name]

    # Create function just like it is done for boundary_function
    name = "solution_function"
    define_boundary_function(solution_expression, name)

    return name


@preamble
def interpolate_vector(name, formdata):
    define_vector(name, formdata)
    element = _driver_data['form'].coefficients()[0].ufl_element()
    bf = name_boundary_function(element)
    gfs = name_gfs(element)
    return "Dune::PDELab::interpolate({}, {}, {});".format(bf,
                                                           gfs,
                                                           name,
                                                           )


@preamble
def interpolate_solution_expression(name):
    formdata = _driver_data['formdata']
    define_vector(name, formdata)
    element = _driver_data['form'].coefficients()[0].ufl_element()
    sol = name_solution_function(element)
    gfs = name_gfs(element)
    return "Dune::PDELab::interpolate({}, {}, {});".format(sol,
                                                           gfs,
                                                           name,
                                                           )


def maybe_interpolate_vector(name, formdata):
    element = _driver_data['form'].coefficients()[0].ufl_element()
    if has_constraints(element):
        interpolate_vector(name, formdata)
    else:
        define_vector(name, formdata)


@symbol
def name_vector(formdata):
    name = form_name_suffix("x", formdata)
    maybe_interpolate_vector(name, formdata)
    return name


@symbol
def name_vector_from_solution_expression():
    interpolate_solution_expression("solution")
    return "solution"


@preamble
def typedef_linearsolver(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::ISTLBackend_SEQ_UMFPack {};".format(name)


@symbol
def type_linearsolver():
    typedef_linearsolver("LinearSolver")
    return "LinearSolver"


@preamble
def define_linearsolver(name):
    lstype = type_linearsolver()
    return "{} {}(false);".format(lstype, name)


@symbol
def name_linearsolver():
    define_linearsolver("ls")
    return "ls"


@preamble
def define_reduction(name):
    ini = name_initree()
    return "double {} = {}.get<double>(\"reduction\", 1e-12);".format(name, ini)


@symbol
def name_reduction():
    define_reduction("reduction")
    return "reduction"


@preamble
def typedef_stationarylinearproblemsolver(name):
    include_file("dune/pdelab/stationary/linearproblem.hh", filetag="driver")
    gotype = type_gridoperator(_driver_data['formdata'])
    lstype = type_linearsolver()
    xtype = type_vector(_driver_data['formdata'])
    return "typedef Dune::PDELab::StationaryLinearProblemSolver<{}, {}, {}> {};".format(gotype, lstype, xtype, name)


@symbol
def type_stationarylinearproblemsolver():
    typedef_stationarylinearproblemsolver("SLP")
    return "SLP"


@preamble
def define_stationarylinearproblemsolver(name):
    slptype = type_stationarylinearproblemsolver()
    formdata = _driver_data['formdata']
    go = name_gridoperator(formdata)
    ls = name_linearsolver()
    x = name_vector(formdata)
    red = name_reduction()
    return "{} {}({}, {}, {}, {});".format(slptype, name, go, ls, x, red)


@symbol
def name_stationarylinearproblemsolver():
    define_stationarylinearproblemsolver("slp")
    return "slp"


@preamble
def typedef_stationarynonlinearproblemsolver(name, go_type):
    include_file("dune/pdelab/newton/newton.hh", filetag="driver")
    ls_type = type_linearsolver()
    x_type = type_vector(_driver_data['formdata'])
    return "typedef Dune::PDELab::Newton<{}, {}, {}> {};".format(go_type, ls_type, x_type, name)


@symbol
def type_stationarynonlinearproblemssolver(go_type):
    typedef_stationarynonlinearproblemsolver("SNP", go_type)
    return "SNP"


@preamble
def define_stationarynonlinearproblemsolver(name, go_type, go):
    snptype = type_stationarynonlinearproblemssolver(go_type)
    x = name_vector(_driver_data['formdata'])
    ls = name_linearsolver()
    return "{} {}({}, {}, {});".format(snptype, name, go, x, ls)


@symbol
def name_stationarynonlinearproblemsolver(go_type, go):
    define_stationarynonlinearproblemsolver("snp", go_type, go)
    return "snp"


@preamble
def typedef_timesteppingmethod(name):
    r_type = type_range()
    explicit = get_option('explicit_time_stepping')
    if explicit:
        return "typedef Dune::PDELab::ExplicitEulerParameter<{}> {};".format(r_type, name)
    else:
        return "typedef Dune::PDELab::OneStepThetaParameter<{}> {};".format(r_type, name)


@symbol
def type_timesteppingmethod():
    typedef_timesteppingmethod("TSM")
    return "TSM"


@preamble
def define_timesteppingmethod(name):
    tsm_type = type_timesteppingmethod()
    explicit = get_option('explicit_time_stepping')
    if explicit:
        return "{} {};".format(tsm_type, name)
    else:
        return "{} {}(1.0);".format(tsm_type, name)


@symbol
def name_timesteppingmethod():
    define_timesteppingmethod("tsm")
    return "tsm"


@preamble
def typedef_instationarygridoperator(name):
    include_file("dune/pdelab/gridoperator/onestep.hh", filetag="driver")
    go_type = type_gridoperator(_driver_data['formdata'])
    mass_go_type = type_gridoperator(_driver_data['mass_formdata'])
    explicit = get_option('explicit_time_stepping')
    if explicit:
        return "typedef Dune::PDELab::OneStepGridOperator<{},{},false> {};".format(go_type, mass_go_type, name)
    else:
        return "typedef Dune::PDELab::OneStepGridOperator<{},{}> {};".format(go_type, mass_go_type, name)


@symbol
def type_instationarygridoperator():
    typedef_instationarygridoperator("IGO")
    return "IGO"


@preamble
def define_instationarygridoperator(name):
    igo_type = type_instationarygridoperator()
    go = name_gridoperator(_driver_data['formdata'])
    mass_go = name_gridoperator(_driver_data['mass_formdata'])
    return "{} {}({}, {});".format(igo_type, name, go, mass_go)


@symbol
def name_instationarygridoperator():
    define_instationarygridoperator("igo")
    return "igo"


@preamble
def typedef_onestepmethod(name):
    r_type = type_range()
    igo_type = type_instationarygridoperator()
    snp_type = type_stationarynonlinearproblemssolver(igo_type)
    vector_type = type_vector(_driver_data['formdata'])
    return "typedef Dune::PDELab::OneStepMethod<{}, {}, {}, {}, {}> {};".format(r_type, igo_type, snp_type, vector_type, vector_type, name)


@symbol
def type_onestepmethod():
    typedef_onestepmethod("OSM")
    return "OSM"


@preamble
def define_onestepmethod(name):
    ilptype = type_onestepmethod()
    tsm = name_timesteppingmethod()
    igo_type = type_instationarygridoperator()
    igo = name_instationarygridoperator()
    snp = name_stationarynonlinearproblemsolver(igo_type, igo)
    return "{} {}({},{},{});".format(ilptype, name, tsm, igo, snp)


@symbol
def name_onestepmethod():
    define_onestepmethod("osm")
    return "osm"


@preamble
def typedef_explicitonestepmethod(name):
    r_type = type_range()
    igo_type = type_instationarygridoperator()
    ls_type = type_linearsolver()
    vector_type = type_vector(_driver_data['formdata'])
    return "typedef Dune::PDELab::ExplicitOneStepMethod<{}, {}, {}, {}> {};".format(r_type, igo_type, ls_type, vector_type, name)


@symbol
def type_explicitonestepmethod():
    typedef_explicitonestepmethod("EOSM")
    return "EOSM"


@preamble
def define_explicitonestepmethod(name):
    eosm_type = type_explicitonestepmethod()
    tsm = name_timesteppingmethod()
    igo = name_instationarygridoperator()
    ls = name_linearsolver()
    return "{} {}({}, {}, {});".format(eosm_type, name, tsm, igo, ls)


@symbol
def name_explicitonestepmethod():
    define_explicitonestepmethod("eosm")
    return "eosm"


@preamble
def dune_solve():
    # Test if form is linear in ansatzfunction
    linear = is_linear(_driver_data['form'])

    # Test wether we want to do matrix free operator evaluation
    matrix_free = get_option('matrix_free')

    if linear and matrix_free:
        formdata = _driver_data['formdata']
        go = name_gridoperator(formdata)
        x = name_vector(formdata)
        include_file("dune/perftool/matrixfree.hh", filetag="driver")
        return "solveMatrixFree({},{});".format(go, x)
    elif linear and not matrix_free:
        slp = name_stationarylinearproblemsolver()
        return "{}.apply();".format(slp)
    elif not linear and matrix_free:
        raise NotImplementedError("Nonlinear and matrix free is not yet implemented")
    elif not linear and not matrix_free:
        go_type = type_gridoperator(_driver_data['formdata'])
        go = name_gridoperator(_driver_data['formdata'])
        snp = name_stationarynonlinearproblemsolver(go_type, go)
        return "{}.apply();".format(snp)
    else:
        assert False


@preamble
def define_vtkfile(name):
    ini = name_initree()
    include_file("string", filetag="driver")
    return "std::string {} = {}.get<std::string>(\"wrapper.vtkcompare.name\", \"output\");".format(name, ini)


@symbol
def name_vtkfile():
    define_vtkfile("vtkfile")
    return "vtkfile"


@preamble
def compare_dofs():
    v = name_vector(_driver_data['formdata'])
    solution = name_vector_from_solution_expression()
    return ["",
            "// Maximum norm of difference between dof vectors of the",
            "// numerical solution and the interpolation of the exact solution",
            "using Dune::PDELab::Backend::native;",
            "double maxerror(0.0);",
            "for (std::size_t i=0; i<native({}).size(); ++i)".format(v),
            "  if (std::abs(native({})[i]-native({})[i]) > maxerror)".format(v, solution),
            "    maxerror = std::abs(native({})[i]-native({})[i]);".format(v, solution),
            "std::cout << std::endl << \"Maxerror: \" << maxerror << std::endl;",
            "if (maxerror>{})".format(get_option("compare_dofs")),
            "  throw;"]


@preamble
def define_discrete_grid_function(gfs, vector_name, dgf_name):
    return ["typedef Dune::PDELab::DiscreteGridFunction<decltype({}),decltype({})> DGF;".format(gfs, vector_name),
            "DGF {}({},{});".format(dgf_name, gfs, vector_name)]


@symbol
def name_discrete_grid_function(gfs, vector_name):
    dgf_name = vector_name + "_dgf"
    define_discrete_grid_function(gfs, vector_name, dgf_name)
    return dgf_name


@preamble
def compare_L2_squared():
    element = _driver_data['form'].coefficients()[0].ufl_element()
    formdata = _driver_data['formdata']
    v = name_vector(formdata)
    gfs = name_gfs(element)
    vdgf = name_discrete_grid_function(gfs, v)
    solution_function = name_solution_function(element)

    include_file("dune/pdelab/gridfunctionspace/gridfunctionadapter.hh", filetag="driver")
    include_file("dune/pdelab/common/functionutilities.hh", filetag="driver")

    return ["",
            "// L2 error squared of difference between numerical",
            "// solution and the interpolation of exact solution",
            "typedef DifferenceSquaredAdapter<decltype({}),decltype({})> DifferenceSquared;".format(solution_function, vdgf),
            "DifferenceSquared differencesquared({},{});".format(solution_function, vdgf),
            "typename DifferenceSquared::Traits::RangeType l2errorsquared(0.0);",
            "Dune::PDELab::integrateGridFunction(differencesquared,l2errorsquared,10);",
            "std::cout << \"l2errorsquared: \" << l2errorsquared << std::endl;",
            "if (l2errorsquared>{})".format(get_option("compare_l2errorsquared")),
            "  throw;"]


@preamble
def print_residual():
    ini = name_initree()
    formdata = _driver_data['formdata']
    n_go = name_gridoperator(formdata)
    v = name_vector(formdata)
    t_v = type_vector(formdata)

    return ["if ({}.get<bool>(\"printresidual\", false)) {{".format(ini),
            "  using Dune::PDELab::Backend::native;",
            "  {} r({});".format(t_v, v),
            "  r=0.0;",
            "  {}.residual({}, r);".format(n_go, v),
            "  Dune::printvector(std::cout, native(r), \"residual vector\", \"row\");",
            "}"]


@preamble
def print_matrix():
    formdata = _driver_data['formdata']
    ini = name_initree()
    t_go = type_gridoperator(formdata)
    n_go = name_gridoperator(formdata)
    v = name_vector(formdata)
    t_v = type_vector(formdata)

    return ["if ({}.get<bool>(\"printmatrix\", false)) {{".format(ini),
            "  typedef typename {}::Traits::Jacobian M;".format(t_go),
            "  M m({});".format(n_go),
            "  {}.jacobian({},m);".format(n_go, v),
            "  using Dune::PDELab::Backend::native;",
            "  Dune::printmatrix(std::cout, native(m),\"global stiffness matrix\",\"row\",9,1);",
            "}"]


@preamble
def define_gfs_name(element, prefix=()):
    from ufl import MixedElement, VectorElement, TensorElement
    if isinstance(element, (VectorElement, TensorElement)):
        gfs = name_gfs(element)
        return "{}.name(\"data_{}\");".format(gfs, "_".join(str(p) for p in prefix))
    if isinstance(element, MixedElement):
        for (i, el) in enumerate(element.sub_elements()):
            define_gfs_name(el, prefix + (i,))
        return ""
    # This is a scalar leaf
    gfs = name_gfs(element)
    return "{}.name(\"data_{}\");".format(gfs, "_".join(str(p) for p in prefix))


@symbol
def type_predicate():
    include_file("dune/perftool/vtkpredicate.hh", filetag="driver")
    return "CuttingPredicate"


@preamble
def define_predicate(name):
    t = type_predicate()
    return "{} {};".format(t, name)


@symbol
def name_predicate():
    define_predicate("predicate")
    return "predicate"


@preamble
def vtkoutput():
    element = _driver_data['form'].coefficients()[0].ufl_element()
    define_gfs_name(element)
    include_file("dune/pdelab/gridfunctionspace/vtk.hh", filetag="driver")
    vtkwriter = name_vtkwriter()
    gfs = name_gfs(element)
    vec = name_vector(_driver_data['formdata'])
    vtkfile = name_vtkfile()
    predicate = name_predicate()
    dune_solve()
    print_residual()
    print_matrix()

    if get_option("exact_solution_expression"):
        from ufl import MixedElement, VectorElement, TensorElement
        if isinstance(_driver_data['form'].coefficients()[0].ufl_element(), (MixedElement, VectorElement, TensorElement)):
            raise NotImplementedError("Comparing to exact solution is olny implemented for scalar elements.")

        if get_option("compare_dofs"):
            compare_dofs()
        if get_option("compare_l2errorsquared"):
            compare_L2_squared()

    return ["Dune::PDELab::addSolutionToVTKWriter({}, {}, {}, Dune::PDELab::vtk::defaultNameScheme(), {});".format(vtkwriter, gfs, vec, predicate),
            "{}.write({}, Dune::VTK::ascii);".format(vtkwriter, vtkfile)]


@preamble
def define_time(name):
    return "double {} = 0.0;".format(name)


@symbol
def name_time():
    define_time("time")
    return "time"


@preamble
def typedef_vtk_sequence_writer(name):
    include_file("dune/grid/io/file/vtk/vtksequencewriter.hh", filetag="driver")
    gv_type = type_leafview()
    return "typedef Dune::VTKSequenceWriter<{}> {};".format(gv_type, name)


@symbol
def type_vtk_sequence_writer():
    typedef_vtk_sequence_writer("VTKSW")
    return "VTKSW"


@preamble
def define_vtk_sequence_writer(name):
    vtksw_type = type_vtk_sequence_writer()
    vtkw_type = type_vtkwriter()
    vtkw = name_vtkwriter()
    vtkfile = name_vtkfile()
    return "{} {}(std::make_shared<{}>({}), {});".format(vtksw_type, name, vtkw_type, vtkw, vtkfile)


@symbol
def name_vtk_sequence_writer():
    define_vtk_sequence_writer("vtkSequenceWriter")
    return "vtkSequenceWriter"


@preamble
def visualize_initial_condition():
    include_file("dune/pdelab/gridfunctionspace/vtk.hh", filetag="driver")
    vtkwriter = name_vtk_sequence_writer()
    element = _driver_data['form'].coefficients()[0].ufl_element()
    define_gfs_name(element)
    gfs = name_gfs(element)
    vector = name_vector(_driver_data['formdata'])
    predicate = name_predicate()
    time = name_time()
    return ["Dune::PDELab::addSolutionToVTKWriter({}, {}, {}, Dune::PDELab::vtk::defaultNameScheme(), {});".format(vtkwriter, gfs, vector, predicate),
            "{}.write({}, Dune::VTK::appendedraw);".format(vtkwriter, time)]


@preamble
def time_loop():
    ini = name_initree()
    formdata = _driver_data['formdata']
    params = name_parameters(formdata)
    time = name_time()
    expr = _driver_data['form'].coefficients()[0].ufl_element()
    bctype = name_bctype_function(expr)
    gfs = name_gfs(expr)
    cc = name_constraintscontainer(expr)
    vector_type = type_vector(formdata)
    vector = name_vector(formdata)

    # Choose between explicit and implicit time stepping
    explicit = get_option('explicit_time_stepping')
    if explicit:
        osm = name_explicitonestepmethod()
        apply_call = "{}.apply(time, dt, {}, {}new);".format(osm, vector, vector)
    else:
        boundary = name_boundary_function(expr)
        osm = name_onestepmethod()
        apply_call = "{}.apply(time, dt, {}, {}, {}new);".format(osm, vector, boundary, vector)

    # Setup visualization
    visualize_initial_condition()
    vtk_sequence_writer = name_vtk_sequence_writer()

    return ["",
            "double T = {}.get<double>(\"instat.T\", 1.0);".format(ini),
            "double dt = {}.get<double>(\"instat.dt\", 0.1);".format(ini),
            "while (time<T-1e-8){",
            "  // Assemble constraints for new time step",
            "  {}.setTime({}+dt);".format(params, time),
            "  Dune::PDELab::constraints({}, {}, {});".format(bctype, gfs, cc),
            "",
            "  // Do time step",
            "  {} {}new({});".format(vector_type, vector, vector),
            "  {}".format(apply_call),
            "",
            "  // Accept new time step",
            "  {} = {}new;".format(vector, vector),
            "  time += dt;",
            "",
            "  // Output to VTK File",
            "  {}.write({}, Dune::VTK::appendedraw);".format(vtk_sequence_writer, time),
            "}",
            ""]


def solve_instationary():
    # Test if form is linear in ansatzfunction
    linear = is_linear(_driver_data['form'])

    # Test wether we want to do matrix free operator evaluation
    matrix_free = get_option('matrix_free')

    # Create time loop
    if linear and matrix_free:
        assert False
    elif linear and not matrix_free:
        time_loop()
    if not linear and matrix_free:
        assert False
    elif not linear and not matrix_free:
        assert False

    print_residual()
    print_matrix()
    if get_option("exact_solution_expression"):
        from ufl import MixedElement, VectorElement, TensorElement
        if isinstance(_driver_data['form'].coefficients()[0].ufl_element(), (MixedElement, VectorElement, TensorElement)):
            raise NotImplementedError("Comparing to exact solution is olny implemented for scalar elements.")

        if get_option("compare_dofs"):
            compare_dofs()
        if get_option("compare_l2errorsquared"):
            compare_L2_squared()


def generate_driver(formdatas, data):
    # The driver module uses a global dictionary for storing necessary data
    set_driver_data(formdatas, data)

    # The vtkoutput is the generating method that triggers all others.
    # Alternatively, one could use the `solve` method.
    if is_stationary():
        vtkoutput()
    else:
        solve_instationary()

    from dune.perftool.generation import retrieve_cache_items
    from cgen import FunctionDeclaration, FunctionBody, Block, Value
    driver_signature = FunctionDeclaration(Value('void', 'driver'), [Value('int', 'argc'), Value('char**', 'argv')])
    driver_body = Block(contents=[i for i in retrieve_cache_items("preamble", make_generable=True)])
    driver = FunctionBody(driver_signature, driver_body)

    filename = get_option("driver_file")

    from dune.perftool.file import generate_file
    generate_file(filename, "driver", [driver])

    # Reset the caching data structure
    from dune.perftool.generation import delete_cache_items
    delete_cache_items()
